from __future__ import annotations

from typing import TYPE_CHECKING, List, Tuple, Union

import jax.numpy as jnp

import photon_weave.extra.einsum_constructor as ESC

if TYPE_CHECKING:
    from photon_weave.state.base_state import BaseState


def apply_operation_vector(
    state_objs: Union[List[BaseState], Tuple[BaseState, ...]],
    target_states: Union[List[BaseState], Tuple[BaseState, ...]],
    product_state: jnp.ndarray,
    operator: jnp.ndarray,
) -> jnp.ndarray:
    """
    Applies the operation to the state vector

    Parameters
    ----------
    state_objs: Union[List[BaseState],Tuple[BaseState,...]]
        List of all base state objects which are in the product state
    states: Union[List[BaseState],Tuple[BaseState,...]]
        List of the base states on which we want to operate
    product_state: jnp.ndarray
        Product stats (state vector)
    operator: jnp.ndarray
        Operator matrix

    Returns
    -------
    jnp.ndarray
        Modified state vector according to the operator

    Notes
    -----
    Given product state needs to be reordered, so that the target states
    are grouped toghether and their index in the tensor (product state)
    reflects their index in the target states reflects their index in the
    operator. Photon Weave handles this automatically when called from
    within the State Comtainer methods.
    """
    assert isinstance(product_state, jnp.ndarray)
    assert isinstance(operator, jnp.ndarray)
    state_objs = list(state_objs)
    target_states = list(target_states)

    operator_shape = jnp.array([s.dimensions for s in target_states])
    dims = jnp.prod(jnp.array([s.dimensions for s in state_objs]))

    shape = [s.dimensions for s in state_objs]
    shape.append(1)

    assert product_state.shape == (dims, 1)
    assert operator.shape == (jnp.prod(operator_shape), jnp.prod(operator_shape))

    product_state = product_state.reshape(shape)
    operator = operator.reshape((*operator_shape, *operator_shape))

    einsum_o = ESC.apply_operator_vector(state_objs, target_states)

    product_state = jnp.einsum(einsum_o, operator, product_state)

    product_state = product_state.reshape((-1, 1))
    # operator = operator.reshape((dims, dims))

    operator = operator.reshape([jnp.prod(operator_shape)] * 2)

    return product_state


def apply_operation_matrix(
    state_objs: Union[List[BaseState], Tuple[BaseState, ...]],
    target_states: Union[List[BaseState], Tuple[BaseState, ...]],
    product_state: jnp.ndarray,
    operator: jnp.ndarray,
) -> jnp.ndarray:
    """
    Applies the operation to the density matrix

    Parameters
    ----------
    state_objs: Union[List[BaseState],Tuple[BaseState,...]]
        List of all base state objects which are in the product state
    states: Union[List[BaseState],Tuple[BaseState,...]]
        List of the base states on which we want to operate
    product_state: jnp.ndarray
        Product stats (state vector)
    operator: jnp.ndarray
        Operator matrix

    Returns
    -------
    jnp.ndarray
        Modified state vector according to the operator

    Notes
    -----
    Given product state needs to be reordered, so that the target states
    are grouped toghether and their index in the tensor (product state)
    reflects their index in the target states reflects their index in the
    operator. Photon Weave handles this automatically when called from
    within the State Comtainer methods.
    """
    assert isinstance(product_state, jnp.ndarray)
    assert isinstance(operator, jnp.ndarray)
    state_objs = list(state_objs)
    target_states = list(target_states)

    state_objs = list(state_objs)
    target_states = list(target_states)
    operator_shape = jnp.array([s.dimensions for s in target_states])
    dims = jnp.prod(jnp.array([s.dimensions for s in state_objs]))
    shape = [s.dimensions for s in state_objs] * 2

    assert product_state.shape == (dims, dims)

    assert operator.shape == (jnp.prod(operator_shape), jnp.prod(operator_shape))

    product_state = product_state.reshape(shape)
    operator = operator.reshape((*operator_shape, *operator_shape))

    dims = jnp.prod(jnp.array([s.dimensions for s in state_objs]))
    shape = [s.dimensions for s in state_objs] * 2
    shape.append(1)

    einsum_o = ESC.apply_operator_matrix(state_objs, target_states)

    product_state = jnp.einsum(einsum_o, operator, product_state, jnp.conj(operator))

    product_state = product_state.reshape((dims, dims))
    operator = operator.reshape((*operator_shape, *operator_shape))
    return product_state


def apply_kraus_matrix(
    state_objs: Union[List[BaseState], Tuple[BaseState, ...]],
    target_states: Union[List[BaseState], Tuple[BaseState, ...]],
    product_state: jnp.ndarray,
    operators: Union[List[jnp.ndarray], Tuple[jnp.ndarray, ...]],
) -> jnp.ndarray:
    """
    Applies the channel described with Kraus operators to the density matrix

    Parameters
    ----------
    state_objs: Union[List[BaseState],Tuple[BaseState,...]]
        List of all base state objects which are in the product state
    states: Union[List[BaseState],Tuple[BaseState,...]]
        List of the base states on which we want to operate
    product_state: jnp.ndarray
        Product stats (state vector)
    operators: Union[List[jnp.ndarray],Tuple[jnp.ndarray,...]]
        List of Operator matrix

    Returns
    -------
    jnp.ndarray
        Modified state vector according to the operator

    Notes
    -----
    Given product state needs to be reordered, so that the target states
    are grouped toghether and their index in the tensor (product state)
    reflects their index in the target states reflects their index in the
    operator. Photon Weave handles this automatically when called from
    within the State Comtainer methods.
    """
    assert isinstance(product_state, jnp.ndarray)
    state_objs = list(state_objs)
    target_states = list(target_states)

    operator_shape = jnp.array([s.dimensions for s in target_states])
    dims = jnp.prod(jnp.array([s.dimensions for s in state_objs]))

    shape = [s.dimensions for s in state_objs]

    assert product_state.shape == (dims, dims)
    expected_operator_dims = int(jnp.prod(operator_shape))
    assert all(
        operator.shape == (expected_operator_dims, expected_operator_dims)
        for operator in operators
    )

    product_state = product_state.reshape((*shape, *shape))

    einsum_o = ESC.apply_operator_matrix(state_objs, state_objs)

    resulting_state = jnp.zeros_like(product_state)

    # Create padding elements
    pre_pad = jnp.array([[1]])
    post_pad = jnp.array([[1]])
    # TODO
    # padding_condition = True
    for state in state_objs:
        if state not in target_states:
            if state_objs.index(state) < state_objs.index(target_states[0]):
                pre_pad = jnp.kron(pre_pad, jnp.eye(state.dimensions))
            elif state_objs.index(state) > state_objs.index(target_states[-1]):
                post_pad = jnp.kron(post_pad, jnp.eye(state.dimensions))

    for operator in operators:
        operator = jnp.kron(pre_pad, jnp.kron(operator, post_pad))
        # We need to expand the operator to affect the whole space
        operator = operator.reshape((*shape, *shape))
        resulting_state += jnp.einsum(
            einsum_o, operator, product_state, jnp.conj(operator)
        )

    resulting_state = resulting_state.reshape((dims, dims))

    return resulting_state



import numpy as np

# CONSTANTS
C0 = 299792458  # m/s


# FUNCTIONS
def gaussian(
    t: float, t_a: float, omega: float, mu: float = 0, sigma: float = 1
) -> float:
    # norm = 1/(sigma*np.sqrt(2 * np.pi))
    norm = 1 / (np.sqrt(sigma * np.pi))
    exponent = -((t - t_a - mu) ** 2) / (2 * sigma**2)
    envelope = norm * np.exp(exponent)
    # carrier = np.exp(1j * omega * t)
    return envelope



"""
Fock state
"""

from __future__ import annotations

import uuid
from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple, Union

import jax.numpy as jnp

from photon_weave._math.ops import num_quanta_matrix, num_quanta_vector
from photon_weave.operation import FockOperationType, Operation
from photon_weave.photon_weave import Config

# from photon_weave.state.composite_envelope import CompositeEnvelope
from .base_state import BaseState
from .expansion_levels import ExpansionLevel
from .utils.measurements import measure_matrix, measure_vector
from .utils.operations import apply_operation_matrix, apply_operation_vector
from .utils.routing import route_operation
from .utils.state_transform import state_contract, state_expand

if TYPE_CHECKING:
    from .envelope import Envelope


class Fock(BaseState):
    """
    Fock class

    This class handles the Fock state or points to the
    Envelope or Composite envelope, which holds the state

    Attributes
    ----------
    index: Union[int, Tuple[int]]
        If Fock space is part of a product space index
        holds information about the space and subspace index
        of this state
    dimension: int
        The dimensions of the Hilbert space, can be set or is
        computed on the fly when expanding the state
    label: int
        If expansion level is Label then label holds the state
        (number basis state)
    state_vector: np.array
        If expansion level is Vector then state_vector holds
        the state
    density_matrix: np.array
        If expansion level is Matrix then density_matrix holds
        the state
    envelope: Envelope
        If the state is part of a envelope, the envelope attribute
        holds a reference to the Envelope instance
    expansion_level: ExpansionLevel
        Holds information about the expansion level of this system
    """

    __slots__ = (
        "uid",
        "index",
        "state",
        "dimensions",
        "density_matrix",
        "envelope",
        "expansions",
        "expansion_level",
        "measured",
    )

    def __init__(self, envelope: Optional[Envelope] = None):
        self.uid: uuid.UUID = uuid.uuid4()
        self.index: Optional[Union[int, Tuple[int, int]]] = None
        self.dimensions: int = -1
        self.state: Optional[Union[int, jnp.ndarray]] = 0
        self.envelope: Optional["Envelope"] = envelope
        self._composite_envelope = None
        self.expansion_level: Optional[ExpansionLevel] = ExpansionLevel.Label
        self.measured = False

    def __hash__(self) -> int:
        return hash(self.uid)

    def __eq__(self, other: Any) -> bool:
        """
        Comparison operator for the states, returns True if
        states are expanded to the same level and are not part
        of the product space
        Todo
        ----
        Method should work for a spaces if they do not have equal
        expansion level
        """
        if not isinstance(other, Fock):
            return False
        if isinstance(self.state, int) and isinstance(other.state, int):
            if self.state == other.state:
                return True
        elif isinstance(self.state, jnp.ndarray) and isinstance(
            other.state, jnp.ndarray
        ):
            if self.state.shape == other.state.shape:
                if jnp.allclose(self.state, other.state):
                    return True
        return False

    @route_operation()
    def expand(self) -> None:
        """
        Expands the representation. If the state is stored in
        label then it is expanded to state_vector and if the
        state is in state_vector, then the state is expanded
        to the state_matrix
        """
        assert self.dimensions is not None, "self.dimensions shoul not be None"
        assert self.state is not None
        assert self.expansion_level is not None

        if self.dimensions < 0:
            if isinstance(self.state, int):
                self.dimensions = self.state + 3

        self.state, self.expansion_level = state_expand(
            self.state, self.expansion_level, self.dimensions
        )

    def contract(
        self, final: ExpansionLevel = ExpansionLevel.Label, tol: float = 1e-6
    ) -> None:
        """
        Attempts to contract the representation to the level defined in final argument.

        Parameters
        ----------
        final: ExpansionLevel
            Expected expansion level after contraction
        tol: float
            Tolerance when comparing matrices
        """
        # If state was measured, then do nothing
        if self.measured:
            return

        assert self.dimensions is not None, "self.dimensions shoul not be None"
        assert self.state is not None
        assert self.expansion_level is not None

        success = True
        while self.expansion_level > final and success:
            self.state, self.expansion_level, success = state_contract(
                self.state, self.expansion_level
            )

    def extract(self, index: Union[int, Tuple[int, int]]) -> None:
        """
        This method is called, when the state is
        joined into a product space. Then the
        index is set and the label, density_matrix and
        state_vector is set to None
        """
        self.index = index
        self.state = None

    @property
    def _num_quanta(self) -> int:
        """
        The highest possible measurement outcome.
        returns highest basis with non_zero probability

        Returns:
        --------
        int
            Highest possible measurement outcome, -1 if failed
        """
        if isinstance(self.state, int):
            return self.state
        elif isinstance(self.state, jnp.ndarray):
            if self.state.shape == (self.dimensions, 1):
                return int(num_quanta_vector(self.state))
            elif self.state.shape == (self.dimensions, self.dimensions):
                return int(num_quanta_matrix(self.state))
        elif self.state is None:
            to = self.trace_out()
            assert isinstance(to, jnp.ndarray)
            if to.shape == (self.dimensions, 1):
                return int(num_quanta_vector(to))
            elif to.shape == (self.dimensions, self.dimensions):
                return int(num_quanta_matrix(to))
        return -1

    def set_index(self, minor: int, major: int = -1) -> None:
        """
        Sets the index, when product space is created, or
        manipulated

        Parameters
        ----------
        minor: int
            Minor index show the order of tensoring in the space
        major: int
            Major index points to the product space when it is in
            CompositeEnvelope
        """
        if major >= 0:
            self.index = (major, minor)
        else:
            self.index = minor

    @route_operation()
    def measure(
        self, separate_measurement: bool = False, destructive: bool = True
    ) -> Dict[BaseState, int]:
        """
        Measures the state in the number basis. This Method can be used if the
        state resides in the Envelope or Composite Envelope

        Parameters
        ----------
        destructive: bool
            If False the state won't be destroyed post measurement (removed)
            The state will still be modified due to measurement
        separate_measurement: bool
            If True and the state is part of the composite envelope
            the state won't be removed from the composite

        Returns
        -------
        Dict[BaseState, int]
            Dictionary of outcomes
        """

        outcomes: Dict[BaseState, int]
        match self.expansion_level:
            case ExpansionLevel.Label:
                assert isinstance(self.state, int)
                outcomes = {self: self.state}
            case ExpansionLevel.Vector:
                assert isinstance(self.state, jnp.ndarray)
                outcomes, post_measurement_state = measure_vector(
                    [self], [self], self.state
                )
            case ExpansionLevel.Matrix:
                assert isinstance(self.state, jnp.ndarray)
                outcomes, post_measurement_state = measure_matrix(
                    [self], [self], self.state
                )

        self.state = outcomes[self]
        self.expansion_level = ExpansionLevel.Label

        if destructive:
            self._set_measured()

        # Handle the case where Fock is included in the Envelope
        if self.envelope is not None and not separate_measurement:
            if not self.envelope.polarization.measured:
                out = self.envelope.polarization.measure(
                    separate_measurement=separate_measurement, destructive=destructive
                )
                for m_key, m_value in out.items():
                    outcomes[m_key] = m_value
        return outcomes

    def _set_measured(self, **kwargs: Dict[str, Any]) -> None:
        """
        Destroys the state
        """
        self.measured = True
        self.state = None
        self.index = None
        self.expansion_level = None

    @route_operation()
    def resize(self, new_dimensions: int) -> bool:
        """
        Resizes the space to the new dimensions.
        If the dimensions are more, than the current dimensions, then
        it gets padded. If the dimensions are less then the current
        dimensions, then it checks if it can shrink the space.

        Parameters
        ----------
        new_dimensions: bool
            New dimensions to be set

        Returns
        -------
        bool
            True if the resizing was succesfull
        """
        # from photon_weave.state.envelope import Envelope

        assert isinstance(self.expansion_level, ExpansionLevel)

        if new_dimensions < 1:
            return False

        if self.index is None:
            if self.expansion_level is ExpansionLevel.Label:
                assert isinstance(self.state, int)
                if self.state > new_dimensions - 1:
                    return False
                else:
                    self.dimensions = new_dimensions
            elif self.expansion_level is ExpansionLevel.Vector:
                assert isinstance(self.state, jnp.ndarray)
                if self.dimensions < new_dimensions:
                    padding_rows = max(0, new_dimensions - self.dimensions)
                    self.state = jnp.pad(
                        self.state,
                        ((0, padding_rows), (0, 0)),
                        mode="constant",
                        constant_values=0,
                    )
                    self.dimensions = new_dimensions
                    return True
                num_quanta = num_quanta_vector(self.state)
                if self.dimensions > new_dimensions and num_quanta < new_dimensions + 1:
                    self.state = self.state[:new_dimensions]
                    self.dimensions = new_dimensions
                    return True
                return False
            elif self.expansion_level is ExpansionLevel.Matrix:
                assert isinstance(self.state, jnp.ndarray)
                assert self.state.shape == (self.dimensions, self.dimensions)
                if new_dimensions > self.dimensions:
                    padding_rows = max(0, new_dimensions - self.dimensions)
                    self.state = jnp.pad(
                        self.state,
                        ((0, padding_rows), (0, padding_rows)),
                        mode="constant",
                        constant_values=0,
                    )
                    self.dimensions = new_dimensions
                    return True
                num_quanta = num_quanta_matrix(self.state)
                if new_dimensions < self.dimensions:
                    self.state = self.state[:new_dimensions, :new_dimensions]
                    self.dimensions = new_dimensions
                    return True
        return False

    @route_operation()
    def apply_operation(self, operation: Operation) -> None:
        """
        Applies an operation to the state. If state is in some product
        state, the operator is correctly routed to the specific
        state

        Decorator:
        ----------
        This method is decorated with `route_operation`, which can based on
        `self.index` execute a method with the same name belonging to
        either `Envelope` or `CompositeEnvelope`. It routes the operation
        to the state container which holds state of this container.

        Parameters
        ----------
        operation: Operation
            Operation with operation type: FockOperationType
        """

        assert isinstance(operation._operation_type, FockOperationType)
        assert isinstance(self.expansion_level, ExpansionLevel)
        assert isinstance(operation.required_expansion_level, ExpansionLevel)

        while self.expansion_level < operation.required_expansion_level:
            self.expand()

        # Consolidate the dimensions
        to = self.trace_out()
        operation.compute_dimensions(self._num_quanta, to)
        self.resize(operation.dimensions[0])
        assert isinstance(self.state, jnp.ndarray)

        match self.expansion_level:
            case ExpansionLevel.Vector:
                self.state = apply_operation_vector(
                    [self], [self], self.state, operation.operator
                )
            case ExpansionLevel.Matrix:
                self.state = apply_operation_matrix(
                    [self], [self], self.state, operation.operator
                )

        if not jnp.any(jnp.abs(self.state) > 0):
            raise ValueError(
                "The state is entirely composed of zeros, is |0⟩"
                " attempted to be annihilated?"
            )
        if operation.renormalize:
            self.state = self.state / jnp.linalg.norm(self.state)

        C = Config()
        if C.contractions:
            self.contract()



import sys
from abc import ABC, abstractmethod
from enum import Enum
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union
from uuid import UUID, uuid4

import jax.numpy as jnp

from photon_weave._math.ops import kraus_identity_check
from photon_weave.photon_weave import Config
from photon_weave.state.expansion_levels import ExpansionLevel

from .utils.measurements import measure_POVM_matrix
from .utils.operations import apply_kraus_matrix
from .utils.representation import representation_matrix, representation_vector
from .utils.routing import route_operation

if TYPE_CHECKING:
    from photon_weave.state.composite_envelope import CompositeEnvelope
    from photon_weave.state.envelope import Envelope
    from photon_weave.state.polarization import PolarizationLabel


class BaseState(ABC):
    __slots__ = (
        "label",
        "_uid",
        "_expansion_level",
        "_index",
        "_dimensions",
        "_measured",
        "_composite_envelope",
        "state",
        "_envelope",
    )

    @abstractmethod
    def __init__(self) -> None:
        self._uid: Union[str, UUID] = uuid4()
        self._expansion_level: Optional[ExpansionLevel] = None
        self._index: Optional[Union[int, Tuple[int, int]]] = None
        self._dimensions: int = -1
        self._composite_envelope: Optional[CompositeEnvelope] = None
        self._envelope: Optional[Envelope] = None
        self.state: Optional[Union[int, PolarizationLabel, jnp.ndarray]] = None

    @property
    def size(self) -> int:
        """
        Returns the size of the state in bytes

        Returns
        -------
        int
            Size of the state in bytes
        """
        if self.state:
            if isinstance(self.state, jnp.ndarray):
                return self.state.nbytes
            else:
                return sys.getsizeof(self.state)
        return 0

    @property
    def envelope(self) -> Union[None, "Envelope"]:
        return self._envelope

    @envelope.setter
    def envelope(self, envelope: Union[None, "Envelope"]) -> None:
        self._envelope = envelope

    @property
    def measured(self) -> bool:
        return self._measured

    @measured.setter
    def measured(self, measured: bool) -> None:
        self._measured = measured

    @property
    def composite_envelope(self) -> Union[None, "CompositeEnvelope"]:
        return self._composite_envelope

    @composite_envelope.setter
    def composite_envelope(
        self, composite_envelope: Union[None, "CompositeEnvelope"]
    ) -> None:
        self._composite_envelope = composite_envelope

    @property
    def uid(self) -> Union[str, UUID]:
        return self._uid

    @uid.setter
    def uid(self, uid: Union[UUID, str]) -> None:
        self._uid = uid

    @property
    def dimensions(self) -> int:
        return self._dimensions

    @dimensions.setter
    def dimensions(self, dimensions: int) -> None:
        self._dimensions = dimensions

    @property
    def expansion_level(self) -> Optional[Union[int, ExpansionLevel]]:
        return self._expansion_level

    @expansion_level.setter
    def expansion_level(self, expansion_level: "ExpansionLevel") -> None:
        self._expansion_level = expansion_level

    @property
    def index(self) -> Union[None, int, Tuple[int, int]]:
        return self._index

    @index.setter
    def index(self, index: Union[None, int, Tuple[int, int]]) -> None:
        self._index = index

    # Dunder methods
    def __hash__(self) -> int:
        return hash(self.uid)

    def __repr__(self) -> str:
        if self.expansion_level == ExpansionLevel.Label:
            if isinstance(self.state, Enum):
                return f"|{self.state.value}⟩"
            elif isinstance(self.state, int):
                return f"|{self.state}⟩"

        if self.index is not None or self.measured:
            return str(self.uid)
        elif self.expansion_level == ExpansionLevel.Vector:
            # Handle cases where the vector has only one element
            assert isinstance(self.state, jnp.ndarray)
            return representation_vector(self.state)
        elif self.expansion_level == ExpansionLevel.Matrix:
            assert isinstance(self.state, jnp.ndarray)
            return representation_matrix(self.state)
        return f"{self.uid}"

    def apply_kraus(
        self,
        operators: List[jnp.ndarray],
        identity_check: bool = True,
    ) -> None:
        """
        Apply Kraus operators to the state.
        State is automatically expanded to the density matrix representation

        Parameters
        ----------
        operators: List[Union[np.ndarray, jnp.Array]]
            List of the operators
        identity_check: bool
            Signal to check whether or not the operators sum up to identity,
            True by default
        """

        assert isinstance(self.expansion_level, ExpansionLevel)

        if identity_check:
            if not kraus_identity_check(operators):
                raise ValueError("Invalid Kraus Channel")
        while self.expansion_level < ExpansionLevel.Matrix:
            self.expand()

        assert isinstance(self.state, jnp.ndarray)
        self.state = apply_kraus_matrix([self], [self], self.state, operators)
 
        C = Config()
        if C.contractions:
            self.contract()

    @abstractmethod
    def expand(self) -> None:
        pass

    @abstractmethod
    def _set_measured(self) -> None:
        pass

    @abstractmethod
    def extract(self, index: Union[int, Tuple[int, int]]) -> None:
        pass

    def contract(
        self, final: ExpansionLevel = ExpansionLevel.Label, tol: float = 1e-6
    ) -> None:
        """
        Attempts to contract the representation to the level defined in final argument.

        Parameters
        ----------
        final: ExpansionLevel
            Expected expansion level after contraction
        tol: float
            Tolerance when comparing matrices
        """

    @route_operation()
    def measure_POVM(
        self,
        operators: List[jnp.ndarray],
        destructive: bool = True,
        partial: bool = False,
    ) -> Tuple[int, Dict["BaseState", int]]:
        """
        Positive Operation-Valued Measurement

        Parameters
        ----------
        operators: List[Union[np.ndarray, jnp.Array]]
            List of POVM operators
        destructive: bool
            If True measurement is destructive and removed (True by default)
        partial: bool
            If True only this state gets measured, if False and part of envelope, the
            corresponding polarization is measured aswell (False by default)

        Returns
        -------
        Tuple[int, Dict[BaseState, int]]
            Returns a tuple, first element is POVM measurement result, second element
            is a dictionary with the other potentional measurement outcomes

        Notes
        -----
        Method is decorated with route_operation. If the state is
        contained in the product state, the corresponding operation
        will be executed in the state container, which contains this
        stat.
        """
        from photon_weave.state.envelope import Envelope
        from photon_weave.state.polarization import Polarization

        assert isinstance(self.expansion_level, ExpansionLevel)
        while self.expansion_level < ExpansionLevel.Matrix:
            self.expand()

        assert isinstance(self.state, jnp.ndarray)
        assert self.state.shape == (self.dimensions, self.dimensions)

        outcome, self.state = measure_POVM_matrix([self], [self], operators, self.state)

        result: Tuple[int, Dict["BaseState", int]] = (outcome, {})
        if destructive:
            self._set_measured()

        if not partial:
            if isinstance(self.envelope, Envelope):
                state = (
                    self.envelope.fock
                    if isinstance(self, Polarization)
                    else self.envelope.polarization
                )
                out = state.measure()
                for k, v in out.items():
                    result[1][k] = v

        C = Config()
        if C.contractions and not destructive:
            self.contract()

        return result

    @route_operation()
    def trace_out(self) -> Union[int, "PolarizationLabel", jnp.ndarray]:
        """
        Returns the traced out state of this base state instance.
        If the instance is in envelope it traces out from there.
        If the instance is in composite envelope then it traces it
        out from there

        Returns
        -------
        Union[int, PolarizationLabel, jnp.ndarray]
            Returns traced out state

        Notes
        -----
        Method is decorated with route_operation. If the state is
        contained in the product state, the corresponding operation
        will be executed in the state container, which contains this
        state.
        """
        assert self.state is not None
        return self.state

    @abstractmethod
    def measure(
        self, separate_measurement: bool = False, destructive: bool = True
    ) -> Dict["BaseState", int]:
        pass



"""
Operations on fock spaces
"""

from enum import Enum
from typing import Any, List, Tuple, Union

import jax.numpy as jnp

from photon_weave._math.ops import (
    annihilation_operator,
    creation_operator,
    displacement_operator,
    phase_operator,
    squeezing_operator,
)
from photon_weave.extra import interpreter
from photon_weave.operation.helpers.fock_dimension_esitmation import FockDimensions
from photon_weave.state.expansion_levels import ExpansionLevel


class FockOperationType(Enum):
    r"""
    FockOperationType

    Constructs an operator, which acts on a single Fock space.

    Creation
    --------
    Constructs a creation operator :math:`\hat a^\dagger`
    Application to a fock state results in
    :math:`\hat a^\dagger |n\rangle = \sqrt{n+1}|n+1\rangle`
    The operator however normalizes the state afterwards, to when using
    Creation operator the resulting state is:
    :math:`\hat a^\dagger |n\rangle = |n+1\rangle`
    The dimensions of the Fock space is changed to the highest number state
    with non zero amplitude plus two.

    Annihilation
    ------------
    Constructs an annihlation operator :math:`\hat a`
    Application to a fock state results in
    :math:`\hat a|n\rangle = \sqrt{n}|n-1\rangle`
    The operator however normalizes the state afterwards, to when using
    Creation operator the resulting state is:
    :math:`\hat a|n\rangle = |n-1\rangle`
    The dimensions of the Fock space is changed to the highest number
    state with non zero amplitude plust two.

    PhaseShift
    ----------
    Constructs a phase shift operator :math:`\hat U(\phi)=e^{-i\phi\hat n}`
    The operation of this operator on a matrix state acts as a identity,
    since global shift is not representable in the density matrix.
    The dimension of the Fock space remains unchanged.

    Squeeze
    -------
    Constructs a Squeezing operator
    :math:`\hat S(\zeta)=e^{\frac{1}{2}(z^*\hat a^2 - z \hat a^\dagger^2)}`
    The dimensions of the Fock space required to accurately represent the application
    of this operator is iteratively determined, resulting in appropriate dimension.

    Displace
    --------
    Constructs a Squeezing operator
    :math:`\hat D(\alpha)=e^{\alpha \hat a^\dagger - \alpha^* \hat a}`
    The dimensions of the Fock space required to accurately represent the application
    of this operator is iteratively determined, resulting in appropriate dimension.
    The dimensions of the Fock space required to accurately represent the application
    of this operator is iteratively determined, resulting in appropriate dimension.

    Identity
    --------
    Constructs a Squeezing operator :math:`\hat I`
    This operator has no effect on the quantum state. The dimensions remain unchanged.

    Custom
    ------
    Constructs a custom operator, which means the operator needs to be manually
    provided, one must pay attention to the fact that operators dimensions need to match
    the dimensions of the Fock space. If the operator is larger than the underlying
    fock space, then the dimensions of the respective Fock space will increase to match
    the provided operator. Otherwise the state will be shrunk. If shrinking is not
    succesfull then the operation will fail. Shinking is not succesfull if part of the
    state is removed by the process of shrinking.

    Usage Example:
    >>> operator = jnp.ndarray(
    >>>    [[0,0],
    >>>     [0,1]]
    >>> )
    >>> op = Operation(FockOperationType.Custom, operator=operator)

    Expresion
    ---------
    Constucts an operator based on the expression provided. Alongside expression
    also list of state types needs to be provided together with the context.
    Context needs to be a dictionary of operators, where the keys are strings used in
    the expression and values are lambda functions, expecting one argument: dimensions.

    An example of context and operator usage
    >>> context = {
    >>>    "a_dag": lambda dims: creation_operator(dims[0])
    >>>    "a":     lambda dims: annihilation_operator(dims[0])
    >>>    "n":     lambda dims: number_operator(dims[0])
    >>> }
    >>> op = Operation(FockOperationType.Expression,
    >>>                expr=("expm"("s_mult", 1j,jnp.pi, "a"))),
    >>>                state_types=(Fock,), # Is applied to only one fock space
    >>>                context=context)
    >>> fock.apply_operation(op)
    """

    Creation: Tuple[bool, List[str], ExpansionLevel, int] = (  # type: ignore[misc]
        True,
        [],
        ExpansionLevel.Vector,
        1,
    )
    Annihilation: Tuple[bool, List[str], ExpansionLevel, int] = (  # type: ignore[misc]
        True,
        [],
        ExpansionLevel.Vector,
        2,
    )
    PhaseShift: Tuple[bool, List[str], ExpansionLevel, int] = (  # type: ignore[misc]
        False,
        ["phi"],
        ExpansionLevel.Vector,
        3,
    )
    Squeeze: Tuple[bool, List[str], ExpansionLevel, int] = (  # type: ignore[misc]
        True,
        ["zeta"],
        ExpansionLevel.Vector,
        4,
    )
    Displace: Tuple[bool, List[str], ExpansionLevel, int] = (  # type: ignore[misc]
        False,
        ["alpha"],
        ExpansionLevel.Vector,
        5,
    )
    Identity: Tuple[bool, List[str], ExpansionLevel, int] = (  # type: ignore[misc]
        False,
        [],
        ExpansionLevel.Vector,
        6,
    )
    Custom: Tuple[bool, List[str], ExpansionLevel, int] = (  # type: ignore[misc]
        False,
        ["operator"],
        ExpansionLevel.Vector,
        7,
    )
    Expresion: Tuple[bool, List[str], ExpansionLevel, int] = (  # type: ignore[misc]
        False,
        ["expr", "context"],
        ExpansionLevel.Vector,
        8,
    )

    def __init__(
        self,
        renormalize: bool,
        required_params: list,
        required_expansion_level: ExpansionLevel,
        op_id: int,
    ) -> None:
        self.renormalize = renormalize
        self.required_params = required_params
        self.required_expansion_level = required_expansion_level

    def update(self, **kwargs: Any) -> None:
        """
        Empty method, doesnt do anything in FockOperationType
        """
        return

    def compute_operator(self, dimensions: List[int], **kwargs: Any) -> jnp.ndarray:
        """
        Generates the operator for this operation, given
        the dimensions

        Parameters
        ----------
        dimensions: List[int]
            The dimensions of the state given in list, for
            this class only one element should be given in list.
            The reason for the list is so that the signature is
            same as in CompositeOperationType
        **kwargs: Any
            List of key word arguments for specific operator types
        """
        match self:
            case FockOperationType.Creation:
                return creation_operator(dimensions[0])
            case FockOperationType.Annihilation:
                return annihilation_operator(dimensions[0])
            case FockOperationType.PhaseShift:
                return phase_operator(dimensions[0], kwargs["phi"])
            case FockOperationType.Displace:
                return displacement_operator(dimensions[0], kwargs["alpha"])
            case FockOperationType.Squeeze:
                return squeezing_operator(dimensions[0], kwargs["zeta"])
            case FockOperationType.Identity:
                return jnp.identity(dimensions[0])
            case FockOperationType.Expresion:
                return interpreter(kwargs["expr"], kwargs["context"], dimensions)
            case FockOperationType.Custom:
                return kwargs["operator"]
        raise ValueError("Something went wrong in operation generation")

    def compute_dimensions(
        self,
        num_quanta: Union[int, List[int]],
        state: jnp.ndarray,
        threshold: float = 1 - 1e-6,
        **kwargs: Any,
    ) -> List[int]:
        """
        Compute the dimensions for the operator. Application of the
        operator could change the dimensionality of the space. For
        example creation operator, would increase the dimensionality
        of the space, if the prior dimensionality doesn't account
        for the new particle.

        Parameters
        ----------
        num_quanta: int
            Number of particles in the space currently. In other words
            highest basis with non_zero probability in the space
        state: jnp.ndarray
            Traced out state, usede for final dimension estimation
        threshold: float
            Minimal amount of state that has to be included in the
            post operation state
        **kwargs: Any
            Additional parameters, used to define the operator

        Returns
        -------
        List[int]
           New number of dimensions in a list for compatibility

        Notes
        -----
        This functionality is called before the operator is computed, so that
        the dimensionality of the space can be changed before the application
        and the dimensionality of the operator and space match
        """
        from photon_weave.operation.operation import Operation

        assert isinstance(num_quanta, int)
        assert isinstance(state, jnp.ndarray)
        state = state / jnp.linalg.norm(state)

        match self:
            case FockOperationType.Creation:
                return [int(num_quanta + 2)]
            case FockOperationType.Annihilation:
                return [num_quanta + 2]
            case FockOperationType.PhaseShift:
                return [num_quanta + 1]
            case FockOperationType.Displace:
                fd = FockDimensions(
                    state,
                    Operation(FockOperationType.Displace, **kwargs),
                    num_quanta,
                    threshold,
                )
                cd = fd.compute_dimensions()
                if cd < num_quanta + 1:
                    cd = num_quanta + 1
                return [cd]
            case FockOperationType.Squeeze:
                fd = FockDimensions(
                    state,
                    Operation(FockOperationType.Squeeze, **kwargs),
                    num_quanta,
                    threshold,
                )
                cd = fd.compute_dimensions()
                if cd < num_quanta + 1:
                    cd = num_quanta + 1
                return [cd]
            case FockOperationType.Identity:
                return [num_quanta + 1]
            case FockOperationType.Expresion:
                fd = FockDimensions(
                    state,
                    Operation(FockOperationType.Expresion, **kwargs),
                    num_quanta,
                    threshold,
                )
                cd = fd.compute_dimensions()
                if cd < num_quanta + 1:
                    cd = num_quanta + 1
                return [cd]
        raise ValueError("Something went wrong in dimension estimation")



from typing import List, Union

import jax.numpy as jnp


def representation_vector(state: jnp.ndarray) -> str:
    """
    Returns the state in a nice string in order to display it

    Parameters
    ----------
    state: jnp.ndarray
        State in the form of the vector

    Returns
    -------
    str:
        String representation of the state vector
    """
    assert isinstance(state, jnp.ndarray)
    assert state.shape[1] == 1

    formatted_vector: Union[str, List[str]]
    formatted_vector = ""
    for row in state:
        formatted_row = "⎢ "
        for num in row:
            # Format real part
            formatted_row += f"{num.real:+.2f} "
            if num.imag >= 0:
                formatted_row += "+ "
            else:
                formatted_row += "- "

            formatted_row += f"{abs(num.imag):.2f}j "

        formatted_row = formatted_row.strip() + " ⎥\n"
        formatted_vector += formatted_row
    formatted_vector = formatted_vector.strip().split("\n")
    formatted_vector[0] = "⎡ " + formatted_vector[0][2:-1] + "⎤"
    formatted_vector[-1] = "⎣ " + formatted_vector[-1][2:-1] + "⎦"
    formatted_vector = "\n".join(formatted_vector)
    return formatted_vector


def representation_matrix(state: jnp.ndarray) -> str:
    """
    Returns the state in a nice string in order to display it

    Parameters
    ----------
    state: jnp.ndarray
        State in the form of a density matrix

    Returns
    -------
    str:
        String representation of the state vector
    """
    assert isinstance(state, jnp.ndarray)
    assert state.shape[0] == state.shape[1]
    formatted_matrix: Union[str, List[str]]
    formatted_matrix = ""

    for row in state:
        formatted_row = "⎢ "
        for num in row:
            formatted_row += f"{num.real:+.2f} "
            if num.imag >= 0:
                formatted_row += "+ "
            else:
                formatted_row += "- "

            # Format the imaginary part and add "j"
            formatted_row += f"{abs(num.imag):.2f}j   "

        formatted_row = formatted_row.strip() + " ⎥\n"
        formatted_matrix += formatted_row

    # Add top and bottom brackets
    formatted_matrix = formatted_matrix.strip().split("\n")
    formatted_matrix[0] = "⎡" + formatted_matrix[0][1:-1] + "⎤"
    formatted_matrix[-1] = "⎣" + formatted_matrix[-1][1:-1] + "⎦"
    formatted_matrix = "\n".join(formatted_matrix)

    return f"{formatted_matrix}"



import itertools
from typing import TYPE_CHECKING, Dict, List

if TYPE_CHECKING:
    from photon_weave.state.base_state import BaseState


def apply_operator_vector(state_objs: list, operator_objs: list) -> str:
    """
    Constructs an Einstein Sum string
    for multiplying an operator with a state vector, where
    operator can be applied to the subspace.

    Parameters
    ----------
    state_objs: List[BaseState]
        List of all State objects which are in the product space
        The order should reflect the order in the tensoring
    operator_objs: List[BaseState]
        List of State objects on which the operator should be applied

    Returns
    -------
    str
        Einstein string that achieves the application

    Notes
    -----
    - The arrays should not be transposed, only reshaped
    """
    einsum_list_list: List[List[int]] = [[], [], []]
    counter = itertools.count(start=0)

    # Consturcting state indices
    einsum_dict: Dict[BaseState, List[int]] = {s: [] for s in state_objs}
    for s in state_objs:
        c = next(counter)
        einsum_dict[s].append(c)
        einsum_list_list[1].append(c)

    # Extra dimension, because vector is still a 2d matrix in jnp view
    ed = next(counter)
    einsum_list_list[1].append(ed)

    # Constructing operation indices
    for s in operator_objs:
        c = next(counter)
        einsum_list_list[0].append(c)
        einsum_dict[s].append(c)

    for s in operator_objs:
        einsum_list_list[0].append(einsum_dict[s][0])

    # Constructing resulting indices
    for s in state_objs:
        if s in operator_objs:
            einsum_list_list[2].append(einsum_dict[s][1])
        else:
            einsum_list_list[2].append(einsum_dict[s][0])
    einsum_list_list[2].append(ed)

    einsum_list_str: List[str] = [
        "".join([chr(97 + i) for i in s]) for s in einsum_list_list
    ]
    return f"{einsum_list_str[0]},{einsum_list_str[1]}->{einsum_list_str[2]}"


def apply_operator_matrix(state_objs: list, operator_objs: list) -> str:
    """
    Constructs an Einstein Sum string
    for multiplying an operator with a density matrix, where
    operator can be applied to the subspace.

    Parameters
    ----------
    state_objs: List[BaseState]
        List of all State objects which are in the product space
        The order should reflect the order in the tensoring
    operator_objs: List[BaseState]
        List of State objects on which the operator should be applied

    Returns
    -------
    str
        Einstein string that achieves the application

    Notes
    -----
    - The arrays should not be transposed, only reshaped

    Examples
    --------
    Suppose you have a state `original_state`, and you want to apply
    an operator to this state. You can reshape the state, apply the
    operator using `einsum`, and then reshape it back to the original
    dimensions.

    Example usage:

    >>> state = original_state.reshape(dims)
    >>> einsum = trace_out(state_objs, operator_objs)
    >>> new_state = jnp.einsum(einsum, operator, state, jnp.conj(operator))
    >>> new_state = new_state.reshape(new_dims)

    In this example:
    - `dims` and `new_dims` are the dimensions of the original and reshaped state.
    - `state_objs` is the list of state objects.
    - `operator_objs` is the list of operator objects.
    - `operator` is the operator to be applied.
    - `einsum` is the Einstein summation string generated by `trace_out`.
    """

    einsum_list_list: List[List[int]] = [[], [], [], []]
    einsum_dict: Dict["BaseState", List[int]] = {k: [] for k in state_objs}
    counter = itertools.count(start=0)

    # Create indices for the current states
    for _ in range(2):
        for s in state_objs:
            c = next(counter)
            einsum_dict[s].append(c)
            einsum_list_list[1].append(c)

    # Create indices for the first operator
    for s in operator_objs:
        c = next(counter)
        einsum_dict[s].append(c)
        einsum_list_list[0].append(c)

    for s in operator_objs:
        einsum_list_list[0].append(einsum_dict[s][0])

    # Create indices for the second operator
    for s in operator_objs:
        c = next(counter)
        einsum_list_list[2].append(c)
        einsum_dict[s].append(c)

    for s in operator_objs:
        einsum_list_list[2].append(einsum_dict[s][1])

    # Create indices for the outcome state

    for s in state_objs:
        if s in operator_objs:
            einsum_list_list[3].append(einsum_dict[s][2])
        else:
            einsum_list_list[3].append(einsum_dict[s][0])

    for s in state_objs:
        if s in operator_objs:
            einsum_list_list[3].append(einsum_dict[s][3])
        else:
            einsum_list_list[3].append(einsum_dict[s][1])

    einsum_list = ["".join([chr(97 + i) for i in s]) for s in einsum_list_list]
    return f"{einsum_list[0]},{einsum_list[1]},{einsum_list[2]}->{einsum_list[3]}"


def trace_out_vector(state_objs: list, states: list) -> str:
    """
    Produces an Einstein sum string. It's application traces out
    the states which are not included in the states list when
    the states are in vector form

    Parameters
    ----------
    state_objs: list
        List of all State objects which are in the product space
        The order should reflect the order in the tensoring
    states: List[BaseState]
        List of State objects which should not be traced out

    Notes
    -----
    - State should not be transposed, only reshaped
    """

    einsum_list_list: List[List[int]] = [[], []]
    c1 = itertools.count(start=0)
    for so in state_objs:
        if so not in states:
            c = next(c1)
            einsum_list_list[0].append(c)
        else:
            c = next(c1)
            einsum_list_list[0].append(c)
            einsum_list_list[1].append(c)
    c = next(c1)
    einsum_list_list[0].append(c)
    einsum_list_list[1].append(c)
    einsum_list = [[chr(97 + x) for x in string] for string in einsum_list_list]
    einsum_str = ["".join(s) for s in einsum_list]

    return f"{einsum_str[0]}->{einsum_str[1]}"


def trace_out_matrix(
    state_objs: list,
    states: list
    ) -> str:
    """
    Produces an Einstein sum string. It's application traces out
    the states which are not included in the states list. Where
    the states are in matrix form.

    Parameters
    ----------
    state_objs: list
        List of all State objects which are in the product space
        The order should reflect the order in the tensoring
    states: List[BaseState]
        List of State objects which should not be traced out

    Notes
    -----
    - State should not be transposed, only reshaped
    """

    einsum_list_list: List[List[int]] = [[], []]
    counter = itertools.count(start=0)
    einsum_dict: Dict["BaseState", int] = {}
    for _ in range(2):
        for so in state_objs:
            if not any(so.uid==s.uid for s in states):
                if so not in einsum_dict.keys():
                    einsum_dict[so] = next(counter)
                einsum_list_list[0].append(einsum_dict[so])
            else:
                c = next(counter)
                einsum_list_list[0].append(c)
                einsum_list_list[1].append(c)

    einsum_list = ["".join([chr(97 + i) for i in s]) for s in einsum_list_list]
    einsum_str = f"{einsum_list[0]}->{einsum_list[1]}"
    return einsum_str


def reorder_vector(state_objs: list, states: list) -> str:
    """
    Produces an Einstein sum string. It's application reorders
    the states in the the product state vector

    Parameters
    ----------
    state_objs: list
        List of all State objects which are in the product space
        The order should reflect the order in the tensoring
    states: List[BaseState]
        States in the new order

    Notes
    -----
    - State should not be transposed, only reshaped
    """
    einsum_dict: Dict["BaseState", int] = {s: -1 for s in state_objs}
    counter = itertools.count(start=0)
    # For the last index
    other = next(counter)
    einsum_list_list: List[List[int]] = [[], []]
    for s in state_objs:
        c = next(counter)
        einsum_list_list[0].append(c)
        einsum_dict[s] = c
    einsum_list_list[0].append(other)

    einsum_list_list[1] = [einsum_dict[s] for s in states]
    einsum_list_list[1].append(other)

    einsum_list = ["".join([chr(97 + s) for s in e]) for e in einsum_list_list]
    return f"{einsum_list[0]}->{einsum_list[1]}"


def reorder_matrix(state_objs: list, states: list) -> str:
    """
    Produces an Einstein sum string. It's application traces out
    the states which are not included in the states list when
    the states are in vector form

    Parameters
    ----------
    state_objs: list
        List of all State objects which are in the product space
        The order should reflect the order in the tensoring
    states: List[BaseState]
        List of State objects which should not be traced out

    Notes
    -----
    - State should not be transposed, only reshaped
    """
    einsum_list_list: List[List[int]] = [[], []]
    einsum_dict: Dict["BaseState", List[int]] = {s: [] for s in state_objs}
    counter = itertools.count(start=0)

    for _ in range(2):
        for s in state_objs:
            c = next(counter)
            einsum_list_list[0].append(c)
            einsum_dict[s].append(c)

    for i in range(2):
        for s in states:
            c = einsum_dict[s][i]
            einsum_list_list[1].append(c)

    einsum_list = ["".join([chr(97 + s) for s in e]) for e in einsum_list_list]
    return f"{einsum_list[0]}->{einsum_list[1]}"


def measure_vector(state_objs: list, states: list) -> str:
    """
    Produces an Einstein sum string. It's application exposes
    the listed states, so they could be measured.

    Parameters
    ----------
    state_objs: list
        List of all State objects which are in the product space
        The order should reflect the order in the tensoring
    states: List[BaseState]
        List of State objects which should be measured

    Notes
    -----
    - State should not be transposed, only reshaped
    """
    einsum_list_list: List[List[int]] = [[], []]
    # einsum_dict: Dict["BaseState", List[int]] = {s: [] for s in state_objs}
    counter = itertools.count(start=0)

    for so in state_objs:
        c = next(counter)
        if so in states:
            einsum_list_list[1].append(c)
        einsum_list_list[0].append(c)

    # Accounting for the verical nature of vector
    c = next(counter)
    einsum_list_list[0].append(c)
    einsum_list_list[1].append(c)

    einsum_list = ["".join([chr(97 + s) for s in e]) for e in einsum_list_list]
    return f"{einsum_list[0]}->{einsum_list[1]}"


def measure_matrix(state_objs: list, states: list) -> str:
    """
    Produces an Einstein sum string. It's application exposes
    the listed states, so they could be measured.

    Parameters
    ----------
    state_objs: list
        List of all State objects which are in the product space
        The order should reflect the order in the tensoring
    states: List[BaseState]
        List of State objects which should be measured

    Notes
    -----
    - State should not be transposed, only reshaped
    """
    einsum_list_list: List[List[int]] = [[], []]
    # einsum_dict: Dict["BaseState", List[int]] = {s: [] for s in state_objs}
    counter = itertools.count(start=0)

    for _ in range(2):
        for so in state_objs:
            c = next(counter)
            if so in states:
                einsum_list_list[1].append(c)
            einsum_list_list[0].append(c)

    einsum_list = ["".join([chr(97 + s) for s in e]) for e in einsum_list_list]
    return f"{einsum_list[0]}->{einsum_list[1]}"



"""
Polarization State
"""

from __future__ import annotations

import logging
import uuid
from enum import Enum
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union

import jax.numpy as jnp

from photon_weave.operation import PolarizationOperationType
from photon_weave.photon_weave import Config

from .base_state import BaseState
from .expansion_levels import ExpansionLevel
from .utils.measurements import measure_matrix, measure_vector
from .utils.operations import apply_operation_matrix, apply_operation_vector
from .utils.routing import route_operation
from .utils.state_transform import state_contract, state_expand

if TYPE_CHECKING:
    from photon_weave.operation import Operation

    from .composite_envelope import CompositeEnvelope
    from .envelope import Envelope

logger = logging.getLogger()


class PolarizationLabel(Enum):
    """
    Labels for the polarization basis states
    """

    H = "H"
    V = "V"
    R = "R"
    L = "L"
    A = "A"
    D = "D"


class Polarization(BaseState):
    """
    Polarization class

    This class handles the polarization state or points to the
    Envelope or Composite envelope, which holds the state

    Attributes
    ----------
    index: Union[int, Tuple[int]]
        If polarization is part of a product space index
        holds information about the space and subspace index
        of this state
    dimension: int
        The dimensions of the Hilbert space (2)
    label: PolarizationLabel
        If expansion level is Label then label holds the state
    state_vector: np.array
        If expansion level is Vector then state_vector holds
        the state
    density_matrix: np.array
        If expansion level is Matrix then density_matrix holds
        the state
    envelope: Envelope
        If the state is part of a envelope, the envelope attribute
        holds a reference to the Envelope instance
    expansion_level: ExpansionLevel
        Holds information about the expansion level of this system
    measured: bool
        If the state was measured than measured is True
    """

    __slots__ = BaseState.__slots__

    def __init__(
        self,
        polarization: PolarizationLabel = PolarizationLabel.H,
        envelope: Union["Envelope", None] = None,
    ):
        self.uid: uuid.UUID = uuid.uuid4()
        logger.info("Creating polarization with uid %s", self.uid)
        self.index: Optional[Union[int, Tuple[int, int]]] = None
        self.state: Optional[Union[jnp.ndarray, PolarizationLabel]] = polarization
        self._dimensions: int = 2
        self.envelope: Optional["Envelope"] = envelope
        self._expansion_level: Optional[ExpansionLevel] = ExpansionLevel.Label
        self._measured: bool = False
        self._composite_envelope: Optional[CompositeEnvelope] = None

    @property
    def measured(self) -> bool:
        return self._measured

    @measured.setter
    def measured(self, measured: bool) -> None:
        self._measured = measured

    @property
    def dimensions(self) -> int:
        return self._dimensions

    @dimensions.setter
    def dimensions(self, dimensions: int) -> None:
        raise ValueError(
            "Dimensions can not be set for Polarization type, 2 by default"
        )

    @route_operation()
    def expand(self) -> None:
        """
        Expands the representation
        If current representation is label, then it gets
        expanded to state_vector and if it is state_vector
        then it gets expanded to density matrix

        Notes
        -----
        Method is decorated with route_operation. If the state is
        contained in the product state, the corresponding operation
        will be executed in the state container, which contains this
        state.
        """
        if self.expansion_level == ExpansionLevel.Label:
            assert isinstance(self.state, PolarizationLabel)
            vector: List[Union[jnp.ndarray, float, complex]]
            match self.state:
                case PolarizationLabel.H:
                    vector = [1, 0]
                case PolarizationLabel.V:
                    vector = [0, 1]
                case PolarizationLabel.R:
                    # Right circular polarization = (1/sqrt(2)) * (|H⟩ + i|V⟩)
                    vector = [1 / jnp.sqrt(2), 1j / jnp.sqrt(2)]
                case PolarizationLabel.L:
                    # Left circular polarization = (1/sqrt(2)) * (|H⟩ - i|V⟩)
                    vector = [1 / jnp.sqrt(2), -1j / jnp.sqrt(2)]
                case PolarizationLabel.A:
                    vector = [1 / jnp.sqrt(2), -1 / jnp.sqrt(2)]
                case PolarizationLabel.D:
                    vector = [1 / jnp.sqrt(2), 1 / jnp.sqrt(2)]
            self.state = jnp.array(vector, dtype=jnp.complex128)[:, jnp.newaxis]
            self.expansion_level = ExpansionLevel.Vector
        else:
            assert isinstance(self.state, jnp.ndarray)
            assert isinstance(self.expansion_level, ExpansionLevel)
            self.state, self.expansion_level = state_expand(
                self.state, self.expansion_level, self.dimensions
            )

    @route_operation()
    def contract(
        self, final: ExpansionLevel = ExpansionLevel.Label, tol: float = 1e-6
    ) -> None:
        """
        Attempts to contract the representation to the level defined in final argument.

        Parameters
        ----------
        final: ExpansionLevel
            Expected expansion level after contraction
        tol: float
            Tolerance when comparing matrices

        Notes
        -----
        Method is decorated with route_operation. If the state is
        contained in the product state, the corresponding operation
        will be executed in the state container, which contains this
        state.
        """
        # If state was measured, then do nothing
        if self.measured:
            return

        if (
            self.expansion_level is ExpansionLevel.Matrix
            and final < ExpansionLevel.Matrix
        ):
            assert isinstance(self.state, (jnp.ndarray, int))
            self.state, self.expansion_level, success = state_contract(  # type: ignore
                self.state, self.expansion_level
            )
        if (
            self.expansion_level is ExpansionLevel.Vector
            and final < ExpansionLevel.Vector
        ):
            assert isinstance(self.state, jnp.ndarray)
            assert self.state.shape == (self.dimensions, 1)
            if jnp.allclose(self.state, jnp.array([[1], [0]])):
                self.state = PolarizationLabel.H
            elif jnp.allclose(self.state, jnp.array([[0], [1]])):
                self.state = PolarizationLabel.V
            elif jnp.allclose(
                self.state, jnp.array([[1 / jnp.sqrt(2)], [1j / jnp.sqrt(2)]])
            ):
                self.state = PolarizationLabel.R
            elif jnp.allclose(
                self.state, jnp.array([[1 / jnp.sqrt(2)], [-1j / jnp.sqrt(2)]])
            ):
                self.state = PolarizationLabel.L
            elif jnp.allclose(
                self.state, jnp.array([[1 / jnp.sqrt(2)], [-1 / jnp.sqrt(2)]])
            ):
                self.state = PolarizationLabel.A
            elif jnp.allclose(
                self.state, jnp.array([[1 / jnp.sqrt(2)], [1 / jnp.sqrt(2)]])
            ):
                self.state = PolarizationLabel.D
            if isinstance(self.state, PolarizationLabel):
                self.expansion_level = ExpansionLevel.Label

    def extract(self, index: Union[int, Tuple[int, int]]) -> None:
        """
        This method is called, when the state is joined into a product space. Then the
        index is set and the label, density_matrix and state_vector is set to None.

        Parameters:

        index: Union[int, Tuple[int, int]
            Index of the state in the product state
        """
        self.index = index
        self.state = None

    def set_index(self, minor: int, major: int = -1) -> None:
        """
        Sets the index, when product space is created, or
        manipulated

        Parameters
        ----------
        minor: int
            Minor index show the order of tensoring in the space
        major: int
            Major index points to the product space when it is in
            CompositeEnvelope
        """
        if self.state is None:
            if major >= 0:
                self.index = (major, minor)
            else:
                self.index = minor
        else:
            raise ValueError("Polarization state does not seem to be extracted")

    def _set_measured(self) -> None:
        """
        Internal method, called after measurement,
        it will destroy the state.
        """
        self.measured = True
        self.state = None
        self.index = None
        self.expansion_level = None

    @route_operation()
    def measure(
        self, separate_measurement: bool = False, destructive: bool = True
    ) -> Dict[BaseState, int]:
        """
        Measures this state. If the state is not in a product state it will
        produce a measurement, otherwise it will return None.

        Parameters
        ----------
        separate_measurement: bool

        Returns
        -------
        Union[int,None]
            Measurement Outcome

        Notes
        -----
        Method is decorated with route_operation. If the state is
        contained in the product state, the corresponding operation
        will be executed in the state container, which contains this
        state.
        """
        if self.expansion_level == ExpansionLevel.Label:
            self.expand()

        match self.expansion_level:
            case ExpansionLevel.Vector:
                assert isinstance(self.state, jnp.ndarray)
                outcomes, post_measurement_state = measure_vector(
                    [self], [self], self.state
                )
            case ExpansionLevel.Matrix:
                assert isinstance(self.state, jnp.ndarray)
                outcomes, post_measurement_state = measure_matrix(
                    [self], [self], self.state
                )
        # Reconstruct the state post measurement
        if outcomes[self] == 0:
            self.state = PolarizationLabel.H
        elif outcomes[self] == 1:
            self.state = PolarizationLabel.V
        self.expansion_level = ExpansionLevel.Label

        if destructive:
            self._set_measured()

        return outcomes

    @route_operation()
    def apply_operation(self, operation: Operation) -> None:
        """
        Applies an operation to the state. If state is in some product
        state, the operator is correctly routed to the specific
        state

        Parameters
        ----------
        operation: Operation
            Operation with operation type: PolarizationOperationType

        Notes
        -----
        Method is decorated with route_operation. If the state is
        contained in the product state, the corresponding operation
        will be executed in the state container, which contains this
        state.
        """
        assert isinstance(operation._operation_type, PolarizationOperationType)
        assert isinstance(self.expansion_level, ExpansionLevel)

        while self.expansion_level < operation.required_expansion_level:
            self.expand()

        operation.compute_dimensions(0, jnp.array([0]))

        match self.expansion_level:
            case ExpansionLevel.Vector:
                assert isinstance(self.state, jnp.ndarray)
                self.state = apply_operation_vector(
                    [self], [self], self.state, operation.operator
                )
            case ExpansionLevel.Matrix:
                assert isinstance(self.state, jnp.ndarray)
                self.state = apply_operation_matrix(
                    [self], [self], self.state, operation.operator
                )

        assert isinstance(self.state, jnp.ndarray)
        if not jnp.any(jnp.abs(self.state) > 0):
            raise ValueError(
                "The state is entirely composed of zeros, is |0⟩ "
                "attempted to be anniilated?"
            )
        if operation.renormalize:
            self.state = self.state / jnp.linalg.norm(self.state)

        C = Config()
        if C.contractions:
            self.contract()



"""
Operations on Custom States
"""

from enum import Enum
from typing import Any, List, Union

import jax.numpy as jnp

from photon_weave.extra import interpreter
from photon_weave.state.expansion_levels import ExpansionLevel


class CustomStateOperationType(Enum):
    """
    CustomStateOperationType

    Expression
    ----------
    Construcs an operator based on the expression provided.
    Alongside expression also a context needs to be provided.
    Context needs to be a dictionary of operators, where the keys are strings
    used in the expression and values are lambda functions, expecting one argument
    dimension. In this case the lambda can also not use the dims parameter,
    but it needs to be defined. Keep in mind that the resulting operator needs
    to have the same dimensionality of the state it is designed to operate on.

    An example of usage:
    >>> context = {
    >>>     "a": lambda dims: jnp.array([[0,0,0],[1,0,0],[0,0,0]]),
    >>>     "b": lambda dims: jnp.array([[0,0,0],[0,0,0],[0,1,0]])
    >>> }
    >>> expr = ("add", "a", "b")
    >>> op = Operation(CustomStateOperationType.Expresion, expr=expr, context=context)

    Custom
    ------
    Constructs a custom operator, which means that the oeprator needs to be manually
    provided. The dimensions of the operator must match the dimensions of the state,
    on which the operator is operating.

    An example of usage:
    >>> operator = jnp.array(
    >>>     [[0,0,0],
    >>>      [1,0,0],
    >>>      [0,1,0]]
    >>> )
    >>> op = Operation(CustomStateOperationType.Custom, operator=operator)
    """

    Expresion = (True, ["expr", "context"], ExpansionLevel.Vector, 1)
    Custom = (True, ["operator"], ExpansionLevel.Vector, 2)

    def __init__(
        self,
        renormalize: bool,
        required_params: list,
        required_expansion_level: ExpansionLevel,
        op_id: int,
    ) -> None:
        self.renormalize = renormalize
        self.required_params = required_params
        self.required_expansion_level = required_expansion_level

    def update(self, **kwargs: Any) -> None:
        """
        Empty method, doesn't do anything in FockOperationType
        """
        return

    def compute_operator(self, dimensions: List[int], **kwargs: Any) -> jnp.ndarray:
        """
        Generates the operator for this operation

        Parameters
        ----------
        dimensions: List[int]
            The dimensions of the sstate given in list, for this
            class only one integer should be given in list.
        **kwargs: Any
            Key word arguments for the operator creation
        """
        match self:
            case CustomStateOperationType.Custom:
                return kwargs["operator"]
            case CustomStateOperationType.Expresion:
                return interpreter(kwargs["expr"], kwargs["context"], dimensions)
        raise ValueError("Operator not recognized")

    def compute_dimensions(
        self,
        num_quanta: Union[int, List[int]],
        state: Union[jnp.ndarray, List[jnp.ndarray]],
        threshold: float = 1,
        **kwargs: Any,
    ) -> List[int]:
        """
        Compute the dimensions for the operator. Since this operator operates
        on a Custom State, which doesn't change dimensions, it always returns
        the dimensions of the given state

        Parameters
        ----------
        num_quanta: Union[int, List[int]]
            Does not have any effect on the output
        state: Union[jnp.ndarray, List[jnp.ndarray]]
            Traced out state, dimensions are computed from this state
        threshold: float
            Does not have any effect on the output
        **kwargs: Any
            Does not have any effect on the output

        Returns
        -------
        List[int]
            List of dimensions with only one element (state.shape[0])
        """
        assert isinstance(state, jnp.ndarray)
        return [state.shape[0]]



# flake8: noqa

from .composite_operation import CompositeOperationType
from .custom_state_operation import CustomStateOperationType
from .fock_operation import FockOperationType
from .operation import Operation
from .polarization_operation import PolarizationOperationType



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable

if TYPE_CHECKING:
    from photon_weave.stats.base_state import BaseState


def route_operation() -> Callable:
    """
    A decorator, which dynamically calls the
    method of the same name of the object, in
    which the state of the caller resides
    """

    def decorator(method: Callable) -> Callable:
        def wrapper(self: BaseState, *args: Any, **kwargs: Any) -> Callable:
            mn = method.__name__

            # Special cases
            if mn == "resize":
                mn = "resize_fock"

            if isinstance(self.index, int):
                if not hasattr(self.envelope, mn):
                    raise AttributeError(f"Envelope has no method {mn}")

                delegate_method = getattr(self.envelope, mn)
                if len(args) > 0:
                    return delegate_method(*args, self, **kwargs)
                else:
                    return delegate_method(self, *kwargs)
            elif isinstance(self.index, tuple):
                if not hasattr(self.composite_envelope, mn):
                    raise AttributeError(f"CompositeEnvelope has no method {mn}")

                delegate_method = getattr(self.composite_envelope, mn)
                if len(args) > 0:
                    return delegate_method(*args, self, **kwargs)
                else:
                    return delegate_method(self, **kwargs)

            # State is in the object
            return method(self, *args, **kwargs)

        return wrapper

    return decorator



class NotExtractedException(Exception):
    pass


class EnvelopeAssignedException(Exception):
    pass


class EnvelopeAlreadyMeasuredException(Exception):
    pass


class MissingTemporalProfileArgumentException(Exception):
    pass
